/**
 * \file sensorManager.hpp
 *
 * Header file for the sensor managers, that decide which data of 
 * which sensor should be integrated in the filter in which order.
 *
 * \date 18/03/2011
 * \author croussil
 *
 * \ingroup rtslam
 */

#ifndef SENSOR_MANAGER_HPP_
#define SENSOR_MANAGER_HPP_

#include "rtslam/sensorAbstract.hpp"

namespace jafar {
namespace rtslam {

	/**
		Base class for sensors managers, that decide which data of which
		sensor should be integrated in the filter next.
		
		The goal is to avoid that a sensor be "starved" (never used)
		because it has large latencies and always arrives too late, once
		the filter has already moved after. In order to do that it can
		use a prediction of the timestamp and arrival time of the next
		data.
		
		The general problem to solve is to chose which updates to do and when,
		in order to maximize the gain of information with the constraints of real time
		computations and chronological updates in filter.
		
		\ingroup rtslam
	*/
	class SensorManagerAbstract
	{
		protected:
			map_ptr_t mapPtr;
			double start_date;
			bool all_init;
		public:
		
		struct ProcessInfo
		{
			sensor_ptr_t sen;
			unsigned id;
			bool no_more_data; // in offline mode, all of the sensors have no more data, so we stop everything
			double date;
			double date_next; // date of next data arrival
			ProcessInfo(sensor_ptr_t sen, unsigned id, double date, double date_next): sen(sen), id(id), no_more_data(false), date(date), date_next(date_next) {}
			//ProcessInfo(sensor_ptr_t sen, unsigned id, double date): sen(sen), id(id), no_more_data(false), date(date), date_next(-1.) {}
			ProcessInfo(bool no_more_data): sen(), id(0), no_more_data(no_more_data), date(-1.), date_next(-1.) {}
			ProcessInfo(): sen(), id(0), no_more_data(false), date(-1.), date_next(-1.) {}
		};
		
		SensorManagerAbstract(map_ptr_t mapPtr): mapPtr(mapPtr), start_date(0.), all_init(false) {}
		
		bool allInit() { return all_init; }
		void setStartDate(double start_date) { this->start_date = start_date; }
		virtual ProcessInfo getNextDataToUse_func(double currentTime) = 0;

		ProcessInfo getNextDataToUse(double currentTime)
		{
			ProcessInfo pinfo;
			while (true)
			{
				pinfo = getNextDataToUse_func(currentTime);
				if (pinfo.sen && !pinfo.sen->getUseForInit() && pinfo.date < start_date)
					pinfo.sen->discard(pinfo.id);
				else
					break;
			}
			return pinfo;
		}
		
		/**
			@return 0 if no more sensor to init, 1 if needs to wait for data to init, 2 if returned correctly a data for init
			*/
		int getNextDataToInit(ProcessInfo &result)
		{
			// TODO set an order for init
			RawInfos infos;
			bool hasSensorMissingData = false;
			sensor_ptr_t oldestSen; unsigned oldestId = 0; double oldestTimestamp = 1e20;
			for (MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin();
				robIter != mapPtr->robotList().end(); ++robIter)
			{
				for (RobotAbstract::SensorList::iterator senIter = (*robIter)->sensorList().begin();
					senIter != (*robIter)->sensorList().end(); ++senIter)
				{
					if ((*senIter)->getUseForInit())
					{
						(*senIter)->queryAvailableRaws(infos);
						if (infos.available.size() > 0)
						{
							RawInfo &info = infos.available.front();
							if (info.timestamp > start_date) // if we anyway don't have data before the start date, use the first one
							{
								if (info.timestamp < oldestTimestamp) { oldestSen = (*senIter); oldestId = info.id; oldestTimestamp = info.timestamp; }
							} else // else find the last one before the start date
							{
								for(std::vector<RawInfo>::reverse_iterator rawIter = infos.available.rbegin(); rawIter != infos.available.rend(); ++rawIter)
									if ((*rawIter).timestamp <= start_date)
									{
										if ((*rawIter).timestamp < oldestTimestamp) { oldestSen = (*senIter); oldestId = (*rawIter).id; oldestTimestamp = (*rawIter).timestamp; }
										break;
									}
							}
						} else
							hasSensorMissingData = true;
					}
				}
			}

			
			if (hasSensorMissingData) return 1; else
			{
				if (!oldestSen) return 0; else
				{
					result = ProcessInfo(oldestSen, oldestId, oldestTimestamp, infos.next.arrival);
					return 2;
				}
			}
		}

		virtual void stop() {}
	};

	
	/**
		This sensors manager is made for offline processing.
		If replay_file is provided, it will integrate data in the same order than
		described by this file.
		If it is not provided, it will integrate all the data, returning at each
		step the oldest unprocessed data.
		
		\ingroup rtslam
	*/
	class SensorManagerOffline: public SensorManagerAbstract
	{
		protected:
			bool replay;
			bool first;
			std::fstream fb;

			size_t sen_id;
			double raw_date;

			bool wait_data;

		public:
			SensorManagerOffline(map_ptr_t mapPtr, std::string replay_path = ""): SensorManagerAbstract(mapPtr)
			{
				wait_data = false;
				first = true;
				if (replay_path.size() != 0)
				{
					replay = true;
					std::string file_dat = replay_path + "/replay.dat";
					fb.open(file_dat.c_str(), std::ios_base::in | std::ios_base::binary);
					if (fb.fail()) { std::cerr << "SensorManagerOffline: could not read file " << file_dat << ", disabling replay" << std::endl; replay = false; }
				}
			}
			
			virtual ProcessInfo getNextDataToUse_func(double currentTime)
			{
				ProcessInfo result;
				RawInfo info;

				if (replay)
				{
					if (first)
					{
						for(MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin();
							robIter != mapPtr->robotList().end(); ++robIter)
						{
							size_t size;
							fb.read((char*)&size, sizeof(size_t));
							if (size != (*robIter)->state.size())
								{ std::cerr << "SensorManager replay: robot state size is " << (*robIter)->state.size() << " but replay file has size " << size << std::endl; return result; }
							for(size_t i = 0; i < size; ++i) fb.read((char*)&((*robIter)->state.x()(i)), sizeof(double));
							for(size_t i = 0; i < size; ++i) for(size_t j = i; j < size; ++j) fb.read((char*)&((*robIter)->state.P()(i,j)), sizeof(double));
							for(size_t i = 0; i < 3; ++i) fb.read((char*)&((*robIter)->start_pos_export(i)), sizeof(double));
							for(size_t i = 0; i < 3; ++i) fb.read((char*)&((*robIter)->start_pos_abs(i)), sizeof(double));
							fb.read((char*)&((*robIter)->is_start_pos_abs_init), sizeof(bool));
						}
						first = false;
					}

					if (!wait_data)
					{
						fb.read((char*)&sen_id, sizeof(size_t));
						fb.read((char*)&raw_date, sizeof(double));
						if (fb.eof()) { result.no_more_data = true; return result; }
					}

					for (MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin();
						robIter != mapPtr->robotList().end(); ++robIter)
					{
						for (RobotAbstract::SensorList::iterator senIter = (*robIter)->sensorList().begin();
							senIter != (*robIter)->sensorList().end(); ++senIter)
						{
							if ((*senIter)->id() == sen_id)
							{
								int status;
								while (true)
								{
									status = (*senIter)->queryNextAvailableRaw(info);
									if (status != 0 || info.timestamp >= raw_date-1e-9) break;
									(*senIter)->discard(info.id);
								}
								if (status == -1) wait_data = true;
								if (status == -2)
								{
									std::cerr << "SensorManagerOffline: missing data from sensor " << sen_id << " at " << std::setprecision(16) << raw_date << std::endl;
									result.no_more_data = true;
								}
								if (status == 0)
								{
									wait_data = false;
									result.id = info.id;
									result.sen = (*senIter);
									result.date = info.timestamp;
								}
								return result;
							}
						}
					}
					std::cerr << "SensorManagerOffline: sensor " << sen_id << " not found !" << std::endl;
					return result;
				} else
				{
					if (!all_init)
					{
						ProcessInfo result;
						int res;
						do {
							res = getNextDataToInit(result);
							if (res == 0) { all_init = true; break; } else
							if (res == 2) { return result; } else
							usleep(1000);
						} while (res == 1);
					}

					result.no_more_data = true;
					wait_data = false;
					for (MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin();
						robIter != mapPtr->robotList().end(); ++robIter)
					{
						for (RobotAbstract::SensorList::iterator senIter = (*robIter)->sensorList().begin();
							senIter != (*robIter)->sensorList().end(); ++senIter)
						{
							if (!(*senIter)->isEnabled()) continue;
							int status = (*senIter)->queryNextAvailableRaw(info);
							if (status == -1) wait_data = true;
							if (status != -2) result.no_more_data = false;
							if (status == 0 && (result.date < 0 || info.timestamp < result.date))
							{
								result.id = info.id;
								result.sen = (*senIter);
								result.date = info.timestamp;
							}
						}
					}

					if (wait_data == true) result.sen.reset();
					return result;
				}
			}
		
			virtual void stop() { if (replay) fb.close(); }
	};



	/**
		This sensor managers tries to integrate several sensors according
		to their policy (all or latest)

		General idea:
		- candidates are the oldest data of "all" and most recent of "latest"
		- smallest next arrival date is limiting

		If data dump is performed and perfect replay may be needed afterwards,
		dump_file should be provided to save which and in which order were data
		integrated.

		\ingroup rtslam
	*/
	class SensorManagerOnline: public SensorManagerAbstract
	{
		protected:
			struct SenInfo
			{
				sensor_ptr_t sen;
				RawInfos infos;
				int res;
				size_t first_ok;
			};
			bool dump;
			bool first;
			std::fstream fl;
			std::fstream fb;
			kernel::LoggerTask *loggerTask;

			class LoggableReplayInitialState: public kernel::Loggable
			{
			 private:
				std::fstream &fl;
				std::fstream &fb;
				jblas::vec stateX;
				jblas::sym_mat stateP;
				jblas::vec start_pos_export;
				jblas::vec start_pos_abs;
				bool is_start_pos_abs_init;

			 public:
				LoggableReplayInitialState(std::fstream &fl, std::fstream &fb, jblas::vec stateX, jblas::sym_mat stateP, jblas::vec start_pos_export, jblas::vec start_pos_abs, bool is_start_pos_abs_init):
					fl(fl), fb(fb), stateX(stateX), stateP(stateP), start_pos_export(start_pos_export), start_pos_abs(start_pos_abs), is_start_pos_abs_init(is_start_pos_abs_init) {}
				virtual void log()
				{
					fl << stateX << std::endl;
					fl << stateP << std::endl;
					fl << start_pos_export << std::endl;
					fl << start_pos_abs << std::endl;
					fl << is_start_pos_abs_init << std::endl;

					size_t size = stateX.size();
					fb.write((char*)&size, sizeof(size_t));
					for(size_t i = 0; i < size; ++i) fb.write((char*)&(stateX(i)), sizeof(double));
					for(size_t i = 0; i < size; ++i) for(size_t j = i; j < size; ++j) fb.write((char*)&(stateP(i,j)), sizeof(double));
					for(size_t i = 0; i < 3; ++i) fb.write((char*)&(start_pos_export(i)), sizeof(double));
					for(size_t i = 0; i < 3; ++i) fb.write((char*)&(start_pos_abs(i)), sizeof(double));
					fb.write((char*)&is_start_pos_abs_init, sizeof(bool));
				}
			};

			class LoggableReplay: public kernel::Loggable
			{
			 private:
				std::fstream &fl;
				std::fstream &fb;
				size_t sen_id;
				std::string sen_name;
				size_t raw_id;
				double raw_date;

			 public:
				LoggableReplay(std::fstream &fl, std::fstream &fb, size_t sen_id, std::string sen_name, size_t raw_id, double raw_date):
					fl(fl), fb(fb), sen_id(sen_id), sen_name(sen_name), raw_id(raw_id), raw_date(raw_date) {}
				virtual void log()
				{
					fl << sen_id << "\t" << sen_name << "\t" << raw_id << "\t" << std::setprecision(19) << raw_date << std::endl;

					fb.write((char*)&sen_id, sizeof(size_t));
					fb.write((char*)&raw_date, sizeof(double));
				}
			};


		public:
			SensorManagerOnline(map_ptr_t mapPtr, std::string dump_path = "", kernel::LoggerTask *loggerTask = NULL):
				SensorManagerAbstract(mapPtr), loggerTask(loggerTask)
			{
				first = true;
				if (dump_path.size() != 0)
				{
					dump = true;
					std::string file_log = dump_path + "/replay.log";
					std::string file_dat = dump_path + "/replay.dat";
					fl.open(file_log.c_str(), std::ios_base::out);
					fb.open(file_dat.c_str(), std::ios_base::out | std::ios_base::binary);
				}
			}

			inline void do_dump(ProcessInfo const &info)
			{
				if (dump) loggerTask->push(new LoggableReplay(fl, fb, info.sen->id(), info.sen->name(), info.id, info.date));
			}

			virtual ProcessInfo getNextDataToUse_func(double currentTime)
			{
				// TODO when we don't receive data from a sensor for some time we should put it in some state
				// stop waiting for these data, and if we receive them again then we start waiting again
				// (eg GPS, or external loc...)

				if (dump && first)
				{
					for(MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin(); robIter != mapPtr->robotList().end(); ++robIter)
						loggerTask->push(new LoggableReplayInitialState(fl, fb, (*robIter)->state.x(), (*robIter)->state.P(), (*robIter)->start_pos_export, (*robIter)->start_pos_abs, (*robIter)->is_start_pos_abs_init));
					first = false;
				}

				// ensure all sensors are init
				if (!all_init)
				{
					ProcessInfo result;
					int res;
					do {
						res = getNextDataToInit(result);
						if (res == 0) { all_init = true; break; } else
						if (res == 2) { do_dump(result); return result; } else
						usleep(1000);
					} while (res == 1);
				}

				// extract information
				std::vector<SenInfo> infos;
				for(MapAbstract::RobotList::iterator robIter = mapPtr->robotList().begin();
					robIter != mapPtr->robotList().end(); ++robIter)
				{
					for(RobotAbstract::SensorList::iterator senIter = (*robIter)->sensorList().begin();
						senIter != (*robIter)->sensorList().end(); ++senIter)
					{
						if (!(*senIter)->isEnabled()) continue;
						SenInfo info;
						info.sen = *senIter;
						info.res = info.sen->queryAvailableRaws(info.infos);
						info.first_ok = 0;

						if (info.infos.integrate_all) // remove data too old to be integrated
						{
							for(; info.first_ok < info.infos.available.size(); ++info.first_ok)
								if (info.infos.available[info.first_ok].timestamp >= currentTime) break;
						} else
						{
							if (info.infos.available.size() > 0 && info.infos.available.back().timestamp < currentTime)
								info.first_ok = info.infos.available.size();
						}

						infos.push_back(info);
					}
				}

				// find largest possible date of data than can be integrated now
				// it is the min of the nexts dates over all the sensors, except if it is late
				double tnow = kernel::Clock::getTime();
				double max_date = tnow + 3600;
				for(std::vector<SenInfo>::iterator it = infos.begin(); it != infos.end(); ++it)
				{
					double next_arrival = it->infos.next.timestamp + (it->infos.next.arrival-it->infos.next.timestamp)*1.2; // add some margin to arrival
					if (next_arrival > tnow && it->infos.next.timestamp < max_date)
						max_date = it->infos.next.timestamp;
				}

				// find the best candidate to integrate now
				// it is the oldest among oldest "all" and newest "latest" that are
				// available and older than max_date
				bool no_more_data = true;
				std::vector<SenInfo>::iterator best_candidate = infos.end();
				RawInfo *best_candidate_info = NULL;
				double oldest_date = tnow + 3600;
				for(std::vector<SenInfo>::iterator it = infos.begin(); it != infos.end(); ++it)
				{
					if (it->res != -2) no_more_data = false;
					if (it->first_ok < it->infos.available.size())
					{
						RawInfo &info = (it->infos.integrate_all ? it->infos.available[it->first_ok] : it->infos.available.back());
						if (info.timestamp < oldest_date) { best_candidate = it; best_candidate_info = &info; oldest_date = info.timestamp; }
					}
				}

				// return result
				if (best_candidate_info)
				{
					ProcessInfo result;
					result = ProcessInfo(best_candidate->sen, best_candidate_info->id, best_candidate_info->timestamp, (best_candidate->infos.integrate_all ? -1 : best_candidate->infos.next.arrival));
					do_dump(result);
					return result;
				}
				else
					return ProcessInfo(no_more_data); // wait
			}

			virtual void stop() { if (loggerTask) loggerTask->join(); if (dump) { fl.close(); fb.close(); } }

	};

	
}}

#endif // SENSORMANAGER_HPP
