/**
 * \file hardwareSensorExternalLoc.cpp
 *
 * File for getting external localization from another agent
 *
 * \date 13/06/2012
 * \author croussil
 *
 * \ingroup rtslam
 */

#include "kernel/timingTools.hpp"
#include "jmath/jblas.hpp"
#include "jmath/ublasExtra.hpp"
#include "rtslam/quatTools.hpp"
#include "rtslam/pinholeTools.hpp"
#include "rtslam/hardwareSensorExternalLoc.hpp"
#include "rtslam/observationPinHoleEuclideanPoint.hpp"

#ifdef HAVE_POSTERLIB
#include "h2timeLib.h"
#endif


namespace jafar {
namespace rtslam {
namespace hardware {


	void HardwareSensorExternalLoc::preloadTask(void)
	{ JFR_GLOBAL_TRY
		ExtLoc data;
#ifdef HAVE_POSTERLIB
		H2TIME h2timestamp;
#endif
		unsigned long prev_ntick = 0;
		double prev_date = 0.0;
		jblas::vec datavec;
		int format, size;

		if (mode == mOnlineDump) log.openWrite("extloc", dump_path, data_type, reading.data.size(), loggerTask);
		if (mode == mOffline)
		{
			log.openRead("extloc", dump_path, format, size);
			ExtLocType actual_data_type = (ExtLocType)format;
			if (actual_data_type != data_type)
			{
				std::cerr << "HardwareSensorExternalLoc: expected type " << data_type << " but got type " << actual_data_type << std::endl;
				return;
			}
		}
		datavec.resize(ExtLocSizes(data_type)+1);
/*
if (mode != 2)
{
sleep(1);
loadACTION(
	"/mnt/data/LAAS/data/2012-09-20_caylus-loc-mana/vol5/test-config.csv",
	"/mnt/data/LAAS/data/2012-09-20_caylus-loc-mana/vol5/test-ObcStates.csv",
	"/mnt/data/LAAS/data/2012-09-20_caylus-loc-mana/vol5/test-ManaDetection.csv",
	"/mnt/data/LAAS/data/2012-09-20_caylus-loc-mana/vol5/test-static.csv");
}
*/
		while (!stopping)
		{
			if (mode == mOffline)
			{
				bool eof = log.read(datavec);
				boost::unique_lock<boost::mutex> l(mutex_data);
				if (isFull(true)) cond_offline_full.notify_all();
				if (eof) { no_more_data = true; cond_offline_full.notify_all(); stopping = true; }
				while (isFull(true) && !stopping) cond_offline_freed.wait(l);
				if (stopping) break;
			} else
			{
#ifdef HAVE_POSTERLIB
				while (!stopping) // wait for new data
				{
					usleep(1000);
					if (posterIoctl(posterId, FIO_GETDATE, &h2timestamp) != ERROR)
					{
						if (h2timestamp.ntick != prev_ntick)
						{
							prev_ntick = h2timestamp.ntick;
							if (posterRead(posterId, 0, &data, sizeof(data)) != ERROR)
							{
								if (data.date != prev_date)
								{
									prev_date = data.date;
									break;
								}
							}
						}
					}
				}
#endif
				reading.arrival = kernel::Clock::getTime();
				if (data_type != data.type)
				{
					std::cout << "HardwareSensorExternalLoc: data type changed from " << data_type << " to " << data.type << ", ignoring this data" << std::endl;
					continue;
				}

				switch (data_type)
				{
					case elCameraBundle: {
						size_t offset = 0;
						datavec(offset++) = data.date;
						for(size_t i = 0; i < 13; ++i) datavec(offset++) = data.meta_x[i];
						for(size_t i = 0; i < 9 ; ++i) datavec(offset++) = data.obs_x[i];
						for(size_t i = 0; i < 13; ++i) datavec(offset++) = data.meta_P[i];
						for(size_t i = 0; i < 32; ++i) datavec(offset++) = data.obs_P[i];
						break;
					}
					case elCameraFullPose: {
						size_t offset = 0;
						datavec(offset++) = data.date;
						for(size_t i = 0; i < 14; ++i) datavec(offset++) = data.meta_x[i];
						for(size_t i = 0; i < 14; ++i) datavec(offset++) = data.obs_x[i];
						for(size_t i = 0; i < 14; ++i) datavec(offset++) = data.meta_P[i];
						for(size_t i = 0; i < 56; ++i) datavec(offset++) = data.obs_P[i];
						break;
					}
					default: {
						std::cerr << "HardwareSensorExternalLoc: data type " << data_type << "not implemented yet." << std::endl;
						break;
					}
				}
			}

			fillReading(datavec);
			
			if (enabled)
			{
				int buff_write = getWritePos(true); // don't need to lock because we are the only writer
				buffer(buff_write) = reading;
				buffer(buff_write).data(0) += timestamps_correction;
				arrival_delay = reading.arrival - buffer(buff_write).data(0);
				last_timestamp = reading.data(0);
				incWritePos();

				if (mode == mOnlineDump) log.write(datavec);
			}
		}

		if (mode == mOnlineDump || mode == mOffline) log.close();

		JFR_GLOBAL_CATCH
	}


	void HardwareSensorExternalLoc::loadACTION(std::string config_filename, std::string state_filename, std::string detect_filename, std::string static_filename)
	{

		double tmpd;
		jblas::vec datavec(ExtLocSizes(elCameraFullPose)+1);

		// read uncert file
		std::fstream static_file; static_file.open(static_filename.c_str(), std::ios_base::in);
		jblas::vec static_line(4), uncertc_line(14), uncerts_line(6), uncertd_line(6);
		dropline(static_file); dropline(static_file); // header lines
		readline(static_file, static_line);
		dropline(static_file); dropline(static_file); // header lines
		readline(static_file, uncertc_line);
		dropline(static_file); dropline(static_file); // header lines
		readline(static_file, uncerts_line);
		dropline(static_file); dropline(static_file); // header lines
		readline(static_file, uncertd_line);
		if (static_line.size() != 4) { std::cerr << "HardwareSensorExternalLoc::loadACTION: Bad format, expected 4 element in " << static_filename << " but got " << static_line.size() << std::endl; return; }
		if (uncertc_line.size() != 14) { std::cerr << "HardwareSensorExternalLoc::loadACTION: Bad format, expected 14 elements in " << static_filename << " but got " << uncertc_line.size() << std::endl; return; }
		if (uncerts_line.size() != 6) { std::cerr << "HardwareSensorExternalLoc::loadACTION: Bad format, expected 6 elements in " << static_filename << " but got " << uncerts_line.size() << std::endl; return; }
		if (uncertd_line.size() != 6 && uncertd_line.size() != 2 && uncertd_line.size() != 0)
			{ std::cerr << "HardwareSensorExternalLoc::loadACTION: Bad format, expected 0 or 2 or 6 elements in " << static_filename << " but got " << uncertd_line.size() << std::endl; return; }

		// read config file
		std::fstream config_file; config_file.open(config_filename.c_str(), std::ios_base::in);
		dropline(config_file); // header line
		config_file >> tmpd; // timestamp
		jblas::vec cameraPose(6);
		config_file >> cameraPose(0); config_file >> cameraPose(1); config_file >> cameraPose(2);
		config_file >> cameraPose(3); config_file >> cameraPose(4); config_file >> cameraPose(5);
		jblas::vec intrinsic(4);
		config_file >> intrinsic(2); config_file >> intrinsic(3); config_file >> intrinsic(0); config_file >> intrinsic(1);
		jblas::vec distortion(3);
		config_file >> distortion(0); config_file >> distortion(1); config_file >> tmpd; config_file >> tmpd; config_file >> distortion(2);

		// read state and detect
		std::fstream state_file; state_file.open(state_filename.c_str(), std::ios_base::in);
		dropline(state_file);
		std::fstream detect_file; detect_file.open(detect_filename.c_str(), std::ios_base::in);
		dropline(detect_file);

		jblas::vec detect_line(13), next_state_line(62), prev_state_line, state_line;

		// preload first two lines
		readline(state_file, next_state_line); prev_state_line = next_state_line;
		bool state_eof = readline(state_file, next_state_line);

		while (!readline(detect_file, detect_line) && !state_eof)
		{
			if (prev_state_line(0) > detect_line(1)) continue; // we don't have loc before this detection
			if (detect_line(2) > 0.01) // != 0 means detection ok
			{
				while (next_state_line(0) < detect_line(1) && !state_eof)
				{
					prev_state_line = next_state_line;
					state_eof = readline(state_file, next_state_line);
				}
				if (next_state_line(0) >= detect_line(1) && !state_eof)
				{
std::cout << "############## load" << std::endl;
					state_line = (prev_state_line*(next_state_line(0)-detect_line(1)) + next_state_line*(detect_line(1)-prev_state_line(0))) / (next_state_line(0)-prev_state_line(0));

					size_t offset = 0;
					reading.arrival = kernel::Clock::getTime()+0.1;
					datavec(offset++) = reading.arrival-0.05; //detect_line(1);
std::cout << "date " << std::setprecision(16) << datavec(offset-1) << std::endl;
					// meta_x
					for(size_t i = 0; i < 4; ++i) datavec(offset++) = intrinsic(i);
					for(size_t i = 0; i < 3; ++i) datavec(offset++) = distortion(i);
					for(size_t i = 0; i < 6; ++i) datavec(offset++) = cameraPose(i);
					datavec(offset++) = static_line(0);
std::cout << "intrinsic " << intrinsic << std::endl;
std::cout << "distortion " << distortion << std::endl;
std::cout << "cameraPose " << cameraPose << std::endl;
std::cout << "physical size " << datavec(offset-1) << std::endl;
					// obs_x
					datavec(offset++) = state_line(11)+static_line(1); // NED to ENU
					datavec(offset++) = state_line(10)+static_line(2);
					datavec(offset++) = -state_line(12)+static_line(3);
std::cout << "ressac pos " << ublas::subrange(datavec, offset-3, offset) << std::endl;
					jblas::vec euler(3);
					for(int i = 0; i < 3; ++i) euler(i) = state_line(13+i);
std::cout << "ressac euler " << euler << std::endl;
					jblas::vec quat = quaternion::e2q(euler);
std::cout << "ressac quat " << quat << std::endl;
					for(int i = 0; i < 4; ++i) datavec(offset++) = quat(i);
					for(size_t i = 0; i < 3; ++i) datavec(offset++) = detect_line(4+i);
std::cout << "detect pos " << ublas::subrange(datavec, offset-3, offset) << std::endl;
					jblas::vec eulerDet(3); eulerDet(0) = 0.0; eulerDet(1) = 0.0; eulerDet(2) = detect_line(8)-M_PI/2;
					if (eulerDet(2) < -M_PI) eulerDet(2) += 2*M_PI; else if (eulerDet(2) > M_PI) eulerDet(2) -= 2*M_PI;
std::cout << "detect euler " << eulerDet << std::endl;
					jblas::vec quatDet = quaternion::e2q(eulerDet);
std::cout << "detect quat " << quatDet << std::endl;
					jblas::vec3 camToRob; camToRob(0) = M_PI; camToRob(1) = 0.0; camToRob(2) = -M_PI/2;
					jblas::vec4 camToRob_q = quaternion::e2q(camToRob);
std::cout << "camToRob quat " << camToRob_q << std::endl;
					ublas::subrange(datavec, offset, offset+4) = quaternion::qProd(quatDet, camToRob_q); offset += 4;
std::cout << "detect quat in rob frame" << ublas::subrange(datavec, offset-4, offset) << std::endl;
					// meta_P
					for(int i = 0; i < 10; ++i) datavec(offset++) = uncertc_line(i)*uncertc_line(i);
					for(int i = 0; i < 3; ++i) datavec(offset++) = uncertc_line(i+10)*M_PI/180.*uncertc_line(i+10)*M_PI/180.;
					datavec(offset++) = uncertc_line(13)*uncertc_line(13);
std::cout << "meta_P " << ublas::subrange(datavec, offset-14, offset) << std::endl;
					// obs_P
					jblas::sym_mat pose_P(7,7); pose_P.clear();
					for(size_t i = 0; i < 3; ++i) pose_P(i,i) = uncerts_line(i)*uncerts_line(i);
					jblas::sym_mat euler_P(3,3); euler_P.clear();
					for(size_t i = 0; i < 3; ++i) euler_P(i,i) = uncerts_line(i+3)*M_PI/180.*uncerts_line(i+3)*M_PI/180.;
					jblas::mat Q_e(4,3); quaternion::e2q_by_de(euler, Q_e);
					ublas::subrange(pose_P, 3,7, 3,7) = prod_JPJt(euler_P, Q_e);
std::cout << "ressac pose_P " << pose_P << std::endl;
					jmath::ublasExtra::explodeSymMat(pose_P, datavec, offset, 7, 0, 7); offset += 7*8/2;
					jblas::sym_mat detect_P(7,7); detect_P.clear();
					jblas::sym_mat eulerDet_P(3,3); eulerDet_P.clear();

					switch (detect_line.size()) // take the uncertainties in detect_line if available, or if uncertd_line otherwise
					{
						case 13:
							for(int i = 0; i < 3; ++i) detect_P(i,i) = detect_line(9+i)*detect_line(9+i);
							eulerDet_P(2,2) = detect_line(12)*detect_line(12);
							break;
						case 9:
							for(int i = 0; i < 3; ++i) detect_P(i,i) = uncertd_line(i)*uncertd_line(i);
							eulerDet_P(2,2) = uncertd_line(3)*M_PI/180.*uncertd_line(3)*M_PI/180.;
							break;
						default:
							std::cerr << "HardwareSensorExternalLoc::loadACTION: Bad format, expected 9 or 13 elements in " << detect_filename << " but got " << detect_line.size() << std::endl;
							continue;
					}
					eulerDet_P(0,0) = uncertd_line(uncertd_line.size()-2)*M_PI/180.*uncertd_line(uncertd_line.size()-2)*M_PI/180.;
					eulerDet_P(1,1) = uncertd_line(uncertd_line.size()-1)*M_PI/180.*uncertd_line(uncertd_line.size()-1)*M_PI/180.;

					quaternion::e2q_by_de(eulerDet, Q_e);
					jblas::mat Q_q(4,4); quaternion::qProd_by_dq1(camToRob_q, Q_q);
					Q_e = ublas::prod(Q_q, Q_e);
					ublas::subrange(detect_P, 3,7, 3,7) = prod_JPJt(eulerDet_P, Q_e);


std::cout << "detect pose_P " << detect_P << std::endl;
					jmath::ublasExtra::explodeSymMat(detect_P, datavec, offset, 7, 0, 7); offset += 7*8/2;

					fillReading(datavec);

					if (enabled)
					{
						int buff_write = getWritePos(true); // don't need to lock because we are the only writer
						buffer(buff_write) = reading;
						buffer(buff_write).data(0) += timestamps_correction;
						arrival_delay = reading.arrival - buffer(buff_write).data(0);
						last_timestamp = reading.data(0);
						incWritePos();
					}

					if (mode == mOnlineDump) log.write(datavec);
				}
			}
		}


	}

	
	void HardwareSensorExternalLoc::fillReading(jblas::vec const & datavec)
	{
		switch (data_type)
		{
			case elCameraBundle: {
// FIXME update, not finished
				// extract data
				double date;
				jblas::vec intrinsic, distortion, correction;
				jblas::vec p, q, o;
				jblas::sym_mat P, Q, O;
				date = datavec(0);
				intrinsic = ublas::subrange(datavec, 1, 5);
				distortion = ublas::subrange(datavec, 5, 8);
				correction.resize(distortion.size()*2);
				pinhole::computeCorrectionModel(intrinsic, distortion, correction);
				p = ublas::subrange(datavec, 8+0, 11);
				q = ublas::subrange(datavec, 8+3, 15);
				o = ublas::subrange(datavec, 8+7, 17);
				P = jmath::ublasExtra::createSymMat(datavec, 17+0, 3, 0, 3);
				Q = jmath::ublasExtra::createSymMat(datavec, 17+6, 4, 0, 4);
				O = jmath::ublasExtra::createSymMat(datavec, 17+16, 2, 0, 2);

				// compute mean of u (standard representation)
				jblas::vec v(3);
				jblas::mat V_o(3,2);
				jblas::mat V_d(3,1);
				pinhole::backProjectPoint(intrinsic, correction, o, 1.0, v, V_o, V_d);
				jblas::vec w(3);
				jblas::mat W_q(3,4);
				jblas::mat W_v(3,3);
				quaternion::rotate(q, v, w, W_q, W_v);
				jblas::vec u = w;
				jblas::mat U_w(3,3);
				jmath::ublasExtra::normalize(u);
				jmath::ublasExtra::normalizeJac(w, U_w);

				// compute covariance matrix of u
				jblas::mat W_o = ublas::prod(W_v, V_o);
				jblas::mat U_o = ublas::prod(U_w, W_o);
				jblas::mat U_q = ublas::prod(U_w, W_q);
				jblas::sym_mat U = jmath::ublasExtra::prod_JPJt(O, U_o) + jmath::ublasExtra::prod_JPJt(Q, U_q);

				// fill standard data vector
				reading.data.clear();
				reading.data(0) = date;
				ublas::subrange(reading.data, 1, 4) = p;
				ublas::subrange(reading.data, 4, 7) = u;
				jmath::ublasExtra::explodeSymMat(P, reading.data, 7, 6, 0, 3);
				jmath::ublasExtra::explodeSymMat(U, reading.data, 13, 6, 3, 3);

				break;
			}
			case elCameraFullPose:
			{
std::cout << "############## fill reading" << std::endl;
				/// extract data
				size_t offset = 0;
				double date = datavec(offset++);
std::cout << "date " << std::setprecision(16) << date << std::endl;
				// meta_x
				jblas::vec intrinsic = ublas::subrange(datavec, offset, offset+4); offset+=4;
				jblas::vec distortion = ublas::subrange(datavec, offset, offset+3); offset+=3;
std::cout << "intrinsic " << intrinsic << std::endl;
std::cout << "distortion " << distortion << std::endl;
				jblas::vec correction(distortion.size()*2); pinhole::computeCorrectionModel(intrinsic, distortion, correction);
				jblas::vec sensorPose(7);
				ublas::subrange(sensorPose, 0, 3) = ublas::subrange(datavec, offset, offset+3); offset+=3;
				jblas::vec euler = ublas::subrange(datavec, offset, offset+3); offset+=3;
				jblas::vec3 trans_e; trans_e(0) = -M_PI/2; trans_e(1) = 0.0; trans_e(2) = -M_PI/2; // transfo from FLU to RBF
std::cout << "cameraPose euler " << euler << std::endl;
				ublas::subrange(sensorPose, 3, 7) = quaternion::qProd(quaternion::e2q(euler), quaternion::e2q(trans_e));
std::cout << "cameraPose " << sensorPose << std::endl;
				double physical_size = datavec(offset++);
std::cout << "physical size " << physical_size << std::endl;
				// obs_x
				jblas::vec robotPose = ublas::subrange(datavec, offset, offset+7); offset+=7;
				jblas::vec impos = ublas::subrange(datavec, offset, offset+2); offset+=2;
				double pix_size = datavec(offset++);
				jblas::vec ori = ublas::subrange(datavec, offset, offset+4); offset+=4;
std::cout << "ressac pose " << robotPose << std::endl;
std::cout << "detect pos " << impos << std::endl;
std::cout << "detect pix size " << pix_size << std::endl;
std::cout << "detect quat " << ori << std::endl;
				// meta_P
				offset+=7; // don't read camera calib variance
				jblas::sym_mat sensorPose_P(7,7); sensorPose_P.clear();
				for(size_t i = 0; i < 3; ++i) sensorPose_P(i,i) = datavec(offset++);
				jblas::sym_mat sensorEuler_P(3,3); sensorEuler_P.clear();
				for(size_t i = 0; i < 3; ++i) sensorEuler_P(i,i) = datavec(offset++);
				jblas::mat Q_e(4,3); quaternion::e2q_by_de(euler, Q_e);
				ublas::subrange(sensorPose_P, 3,7, 3,7) = prod_JPJt(sensorEuler_P, Q_e);
				double physical_size_P = datavec(offset++);
std::cout << "sensorPose_P " << sensorPose_P << std::endl;
std::cout << "physical_size_P " << physical_size_P << std::endl;
				// obs_P
				jblas::sym_mat robotPose_P = jmath::ublasExtra::createSymMat(datavec, offset, 7, 0, 7); offset+=7*8/2;
				jblas::sym_mat detection_P = jmath::ublasExtra::createSymMat(datavec, offset, 7, 0, 7); offset+=7*8/2;
std::cout << "ressac pose_P " << robotPose_P << std::endl;
std::cout << "detect pose_P " << detection_P << std::endl;

				/// compute values
				jblas::mat SG_r(7, 7), SG_s(7, 7);
				jblas::vec sensorGlobalPose(7);
				quaternion::composeFrames(robotPose, sensorPose, sensorGlobalPose, SG_r, SG_s);
				jblas::sym_mat sensorGlobalPose_P = jmath::ublasExtra::prod_JPJt(robotPose_P, SG_r) + jmath::ublasExtra::prod_JPJt(sensorPose_P, SG_s);

				ObservationModelPinHoleEuclideanPoint model;
				double fc = (intrinsic(2)+intrinsic(3))/2; // average focal to compute dist
				jblas::vec idist(1); idist(0) = 1.0 / (fc * physical_size / pix_size); // inverse distance
std::cout << "fc " << fc << " phsize " << physical_size << " psize " << pix_size << " idist " << idist(0) << std::endl;
				jblas::vec pos;
				jblas::mat POS_sg(3,7), POS_p(3,2), POS_d(3,1);
				model.backProject_func_func(intrinsic, correction, quaternion::originFrame(), impos, idist, pos, POS_sg, POS_p, POS_d);
std::cout << "pos after backproject " << pos << std::endl;

				double DEPTH_phs = fc * 1./pix_size;
				double DEPTH_pxs = fc * (-physical_size / (pix_size*pix_size));
				jblas::sym_mat depth_P(1,1); depth_P(0,0) = DEPTH_phs*physical_size_P*DEPTH_phs + DEPTH_pxs*detection_P(2,2)*DEPTH_pxs;
				jblas::sym_mat pos_P = /*jmath::ublasExtra::prod_JPJt(sensorGlobalPose_P, POS_sg) +*/
				                       jmath::ublasExtra::prod_JPJt(ublas::subrange(detection_P,0,2,0,2), POS_p) +
				                       jmath::ublasExtra::prod_JPJt(depth_P, POS_d);

				jblas::vec objectPose(7), objectGlobalPose(7);
				ublas::subrange(objectPose, 0,3) = pos;
				ublas::subrange(objectPose, 3,7) = ori;
std::cout << "sensorGlobalPose " << ublas::subrange(sensorGlobalPose,0,3) << " " << ublas::subrange(sensorGlobalPose,3,7) << " " << quaternion::q2e(ublas::subrange(sensorGlobalPose,3,7)) << std::endl;
std::cout << "objectPose " << ublas::subrange(objectPose,0,3) << " " << quaternion::q2e(ublas::subrange(objectPose,3,7)) << std::endl;
				quaternion::composeFrames(sensorGlobalPose, objectPose, objectGlobalPose, SG_r, SG_s);
				jblas::sym_mat objectPose_P(7,7);
				ublas::subrange(objectPose_P, 0,3, 0,3) = pos_P;
				ublas::subrange(objectPose_P, 3,7, 3,7) = ublas::subrange(detection_P, 3,7, 3,7);
				jblas::sym_mat objectGlobalPose_P = jmath::ublasExtra::prod_JPJt(sensorGlobalPose_P, SG_r) + jmath::ublasExtra::prod_JPJt(objectPose_P, SG_s);

				/// fill standard data vector
				reading.data.clear(); offset = 0;
				reading.data(offset++) = date;
				ublas::subrange(reading.data, offset, offset+7) = ublas::subrange(objectGlobalPose, 0,3); offset+=3;
				jmath::ublasExtra::explodeSymMat(objectGlobalPose_P, reading.data, offset, 7, 0, 3); offset+=3*4/2;
//				ublas::subrange(reading.data, offset, offset+7) = ublas::subrange(objectGlobalPose, 0,7); offset+=7;
//				jmath::ublasExtra::explodeSymMat(objectGlobalPose_P, reading.data, offset, 7, 0, 7); offset+=7*8/2;
std::cout << "observation of pose " << ublas::subrange(objectGlobalPose,0,3) << " " << ublas::subrange(objectGlobalPose,3,7) << " " << quaternion::q2e(ublas::subrange(objectGlobalPose,3,7)) << std::endl;
std::cout << " with P " << objectGlobalPose_P << std::endl;
				break;
			}
			default: {
				std::cout << "hardwareSensorExternalLoc: data type " << data_type << " not managed." << std::endl;
				return;
			}
		}

	}


	
	HardwareSensorExternalLoc::HardwareSensorExternalLoc(kernel::VariableCondition<int> *condition, unsigned bufferSize, const std::string machine, ExtLocType data_type, int source, Mode mode, std::string dump_path, kernel::LoggerTask *loggerTask):
		HardwareSensorProprioAbstract(condition, mode, bufferSize, ctFull), data_type(data_type), source(source), dump_path(dump_path), loggerTask(loggerTask)
	{
		if (mode == mOnlineDump && !loggerTask) JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "HardwareSensorExternalLoc: you must provide a loggerTask if you want to dump data.");
		switch (data_type)
		{
			case elCameraBundle:
				addQuantity(qBundleobs);
				break;
			case elCameraFullPose:
				addQuantity(qPos);
				//addQuantity(qOriQuat);
				break;
			default:
				std::cout << "hardwareSensorExternalLoc: data type " << data_type << " not managed." << std::endl;
				return;
		}
		initData();

		// configure
		if (source == 0 && (mode == mOnline || mode == mOnlineDump))
		{
#ifdef HAVE_POSTERLIB
			char *backup_poster_path = getenv("POSTER_PATH");
			setenv("POSTER_PATH", machine.c_str(), 1);
			if (posterFind("extloc", &this->posterId) == ERROR) {
				JFR_ERROR(RtslamException, RtslamException::GENERIC_ERROR, "Poster extloc could not be found");
			}
			if (backup_poster_path) setenv("POSTER_PATH", backup_poster_path, 1); else unsetenv("POSTER_PATH");
#else
			JFR_ERROR(RtslamException, RtslamException::MISSING_DEPENDENCY,
			          "You must install posterlib (pocolibs) if you want to use online GPS sensor");
#endif
		}
	}

	void HardwareSensorExternalLoc::start()
	{
		// start acquire task
		//preloadTask();
		if (started) { std::cout << "Warning: This HardwareSensorExternalLoc has already been started" << std::endl; return; }
		started = true;
		if (source) stopping = true;
		last_timestamp = kernel::Clock::getTime();
		preloadTask_thread = new boost::thread(boost::bind(&HardwareSensorExternalLoc::preloadTask,this));
		
		if (mode == mOffline)
		{ // wait that log has been read before first frame
			boost::unique_lock<boost::mutex> l(mutex_data);
			cond_offline_full.wait(l);
		}
	}

	void HardwareSensorExternalLoc::stop()
	{
		if (!started) return;
		stopping = true;
		cond_offline_freed.notify_all();
	}

	void HardwareSensorExternalLoc::join()
	{
		if (boost::this_thread::get_id() != preloadTask_thread->get_id())
			preloadTask_thread->join();
	}
	
}}}

