/**
 * \file robotInertial.cpp
 * \date 26/03/2010
 * \author jsola
 * \ingroup rtslam
 */

#include "jmath/jblas.hpp"

//#include "rtslam/robotAbstract.hpp"
#include "rtslam/robotInertial.hpp"
#include "rtslam/sensorAbstract.hpp"

#include "rtslam/quatTools.hpp"

namespace jafar {
	namespace rtslam {
		using namespace std;
		using namespace jblas;
		using namespace ublas;
		using namespace quaternion;


#if ESTIMATE_G_VEC
			int RobotInertial::g_size = 3;
#else
			int RobotInertial::g_size = 1;
#endif


		/**
		 * Remote constructor from remote map.
		 * \param _map the remote map
		 */
		RobotInertial::RobotInertial(const map_ptr_t & _mapPtr):
			RobotAbstract(_mapPtr)
		{
			iaPos = getQuantity(qPos);
			iaOriQuat = getQuantity(qOriQuat);
			registerRobotQuantity(qAbsVel);
			iaAbsVel = getQuantity(qAbsVel);
			iaBiasA = allocateSensorQuantity(3);
			iaBiasW = allocateSensorQuantity(3);
			Quantity quant = (g_size == 1 ? qNormG : qG);
			registerRobotQuantity(quant);
			iaG = getQuantity(quant);

			#if 0
			registerRobotQuantity(qAngVel);
			#endif

			JFR_ASSERT(RobotInertial::size() == state_size, "wrong state size"); // temporary
			RobotAbstract::construct(_mapPtr, state_size, RobotInertial::size_control(), RobotInertial::size_perturbation());

			constantPerturbation = false;
			type = INERTIAL;
			z_axis.clear(); z_axis(2) = 1;
		}



		/*
		 * This motion model is driven by IMU measurements and random perturbations, and defined by:
		 * The state vector, x = [p q v ab wb g] , of size 19.
		 *
		 * The transition equation is
		 * - x+ = move_func(x,u,n),
		 *
		 * with u = [am, wm] the IMU measurements (the control input)
		 *  and n = [vi, ti, abi, wbi] the perturbation impulse.
		 *
		 * the transition equation f() is decomposed as:
		 * #if USE_NEW_STATE
		 * - p+  = p + (v + v+)/2*dt
		 * #else
		 * - p+  = p + v*dt
		 * #endif
		 * - q+  = qnorm(q**((wm - wb)*dt + ti)        <-- ** : quaternion product ; ti : theta impulse
		 * #if USE_NEW_STATE
		 * - v+  = v + (R(q+)*(am - ab) + g)*dt + vi
		 * #else
		 * - v+  = v + (R(q)*(am - ab) + g)*dt + vi    <-- am and wm: IMU measurements ; vi : v impulse
		 * #endif
		 * - ab+ = ab + abi                            <-- abi : random walk in acc bias with abi impulse perturbation
		 * - wb+ = wb + wbi                            <-- wbi : random walk of gyro bias with wbi impulse perturbation
		 * - g+  = g                                   <-- g  : gravity vector, constant but unknown
		 * -----------------------------------------------------------------------------
		 *
		 * The Jacobian XNEW_x is built with
		 *   var    |  p          q              v        ab             wb             g
		 *      pos |  0          3              7        10             13             16
		 *   -------+-------------------------------------------------------------------
		 * #if USE_NEW_STATE
		 *   p   0  |  I  VNEW_q*QNEW_q*dt/2   I*dt  -R*dt*dt/2 VNEW_q*QNEW_wb*dt/2  I*dt*dt/2   (+ dt/2 * VNEW_)
		 *   q   3  |  0        QNEW_q           0        0            QNEW_wb          0        (QNEW_)
		 *   v   7  |  0     VNEW_q*QNEW_q       I      -R*dt       VNEW_q*QNEW_wb     I*dt      (+ VNEW_q * QNEW_)
		 * #else
		 *   p   0  |  I          0             I*dt      0              0              0
		 *   q   3  |  0        QNEW_q           0        0            QNEW_wb          0
		 *   v   7  |  0        VNEW_q           I      -R*dt            0             I*dt
		 * #endif
		 *   ab  10 |  0          0              0        I              0              0
		 *   wb  13 |  0          0              0        0              I              0
		 *   g   16 |  0          0              0        0              0              I
		 * -----------------------------------------------------------------------------
		 *
		 * The Jacobian XNEW_pert is built with
		 *   var    |  vi             ti             abi   wbi
		 *      pos |  0              3              6     9
		 *   -------+-------------------------------------------------------------------
		 * #if USE_NEW_STATE
		 *   p   0  |I.dt/2  VNEW_q*QNEW_ti*dt/2     0     0     (+ dt/2 * VNEW_)
		 *   q   3  |  0            QNEW_ti          0     0     (QNEW_)
		 *   v   7  |  I       VNEW_q*QNEW_ti        0     0     (+ VNEW_q * QNEW_)
		 * #else
		 *   p   0  |  0              0              0     0
		 *   q   3  |  0            QNEW_ti          0     0
		 *   v   7  |  I              0              0     0
		 * #endif
		 *   ab  10 |  0              0              I     0
		 *   wb  13 |  0              0              0     I
		 *   g   16 |  0              0              0     0
		 * -----------------------------------------------------------------------------
		 */
		void RobotInertial::move_func(const vec & _x, const vec & _u, const vec & _n, double _dt, vec & _xnew, mat & _XNEW_x,
				mat & _XNEW_pert, unsigned tempSet) const
		{
			TempVariables & t = (tempSet == 0 ? tempvars0 : tempvars1);
			if (tempSet > 1) std::cerr << "Error: RobotInertial has only 2 sets of temps variables, increase it if you need more" << std::endl;

			// Separate things out to make it clearer
			vec3 p, v, ab, wb, gv;
			vec g;
			vec4 q;
			splitState(_x, p, q, v, ab, wb, g); // split state vector
			if (g_size == 1) { gv = z_axis * g(0); } else { gv = g; }

			// Split control and perturbation vectors into
			// sensed acceleration and sensed angular rate
			// and noises
			vec3 am, wm, vi, ti, abi, wbi; // measurements and random walks
			splitControl(_u, am, wm);
			splitPert(_n, vi, ti, abi, wbi);

			// Get new state vector
			vec3 pnew, vnew, abnew, wbnew;
			vec gnew;
			vec4 qnew;

			// It is useful to start obtaining a nice rotation matrix and the product R*dt
			t.Rdt = q2R(q) * _dt;

			// compute in the order Q,V,P because of dependencies with USE_NEW_STATE

			//-------------------------------- COMPUTE Q
			// Invert sensor functions. Get true angular rates:
			// w = wsens - wb
			vec3 wtrue = wm - wb;

			// qnew = q x q(w * dt) (keep qwdt for later use)
			vec4 qwdt = v2q(wtrue * _dt + ti);
			qnew = qProd(q, qwdt);
			ublasExtra::normalizeJac(qnew, t.QNEW_qn);
			ublasExtra::normalize(qnew);

			//-------------------------------- COMPUTE V
			// Invert sensor functions. Get true acceleration:
			// a = R(q)(asens - ab) + g
			#if USE_NEW_STATE
			vec3 atrue = rotate(qnew, (am - ab)) + gv;
			#else
			vec3 atrue = rotate(q, (am - ab)) + gv;
			#endif
			vnew = v + atrue * _dt + vi; //    velocity

			//-------------------------------- COMPUTE P
			#if USE_NEW_STATE
			pnew = p + (v+vnew)/2 * _dt; //     position
			#else
			pnew = p + v * _dt; //     position
			#endif

			//-------------------------------- COMPUTE biases
			abnew = ab + abi; //          acc bias
			wbnew = wb + wbi; //          gyro bias
			gnew = g; //                 gravity does not change
			
			// Put it all together - this is the output state
			unsplitState(pnew, qnew, vnew, abnew, wbnew, gnew, _xnew);


		/* Now on to the Jacobian XNEW_x...
		 * Identity is a good place to start since overall structure is like this
		 *   var    |  p          q              v        ab             wb             g
		 *      pos |  0          3              7        10             13             16
		 *   -------+-------------------------------------------------------------------
		 * #if USE_NEW_STATE
		 *   p   0  |  I  VNEW_q*QNEW_q*dt/2   I*dt  -R*dt*dt/2 VNEW_q*QNEW_wb*dt/2  I*dt*dt/2   (+ dt/2 * VNEW_)
		 *   q   3  |  0        QNEW_q           0        0            QNEW_wb          0        (QNEW_)
		 *   v   7  |  0     VNEW_q*QNEW_q       I      -R*dt       VNEW_q*QNEW_wb     I*dt      (+ VNEW_q * QNEW_)
		 * #else
		 *   p   0  |  I          0             I*dt      0              0              0
		 *   q   3  |  0        QNEW_q           0        0            QNEW_wb          0
		 *   v   7  |  0        VNEW_q           I      -R*dt            0             I*dt
		 * #endif
		 *   ab  10 |  0          0              0        I              0              0
		 *   wb  13 |  0          0              0        0              I              0
		 *   g   16 |  0          0              0        0              0              I
		 */

			_XNEW_x.assign(identity_mat(state.size()));

			//-------------------------------- COMPUTE P

			// Fill in PNEW_v
			identity_mat I(3);
			t.Idt = I * _dt;
			mat Iz; if (g_size == 1) { Iz.resize(3,1); Iz.clear(); Iz(2,0)=1; } else { Iz = I; }
			mat Izdt = Iz * _dt;
			project(_XNEW_x, iaPos, iaAbsVel) = t.Idt;
			#if USE_NEW_STATE
			// true idea, but not efficient, instead of previous line:
			//subrange(_XNEW_x, iaPos, iaAbsVel) = t.Idt/2; // because uses v/2, but another Idt/2 would come from v+ dependency
			#endif

			//-------------------------------- COMPUTE V

			// Fill VNEW_q
			// VNEW_q = d(R(q)*v) / dq
			rotate_by_dq(q, am-ab, t.VNEW_q); t.VNEW_q *= _dt;
			#if !USE_NEW_STATE
			subrange(_XNEW_x, 7, 10, 3, 7) = t.VNEW_q;
			#else
			// doesn't depend directly of q anymore, everything will come from q+ dependency
			#endif

			// Fill in VNEW_ab
			project(_XNEW_x, iaAbsVel, iaBiasA) = -t.Rdt;

			// Fill in VNEW_g
			project(_XNEW_x, iaAbsVel, iaG) = Izdt;

			//-------------------------------- COMPUTE Q

			// Fill in QNEW_q
			// qnew = qold ** qwdt  ( qnew = q1 ** q2 = qProd(q1, q2) in rtslam/quatTools.hpp )
			qProd_by_dq1(qwdt, t.QN_q);
			project(_XNEW_x, iaOriQuat, iaOriQuat) = prod(t.QNEW_qn, t.QN_q);

			// Fill in QNEW_wb
			// QNEW_wb = QNEW_qwdt * QWDT_wdt * WDT_w * W_wb
			//         = QNEW_qwdt * QWDT_w * W_wb
			//         = QNEW_qwdt * QWDT_w * (-1)
			qProd_by_dq2(q, t.QN_qwdt);
			// Here we get the derivative of qwdt wrt wtrue, so we consider dt = 1 and call for the derivative of v2q() with v = w*dt
//			v2q_by_dv(wtrue, QWDT_w); // TODO check equivalent
			v2q_by_dv(wtrue*_dt, t.QWDT_w); t.QWDT_w *= _dt;
			t.QN_w = prod(t.QN_qwdt, t.QWDT_w);
			t.QNEW_w = prod(t.QNEW_qn, t.QN_w);
			project(_XNEW_x, iaOriQuat, iaBiasW) = -t.QNEW_w;


			//-------------------------------- COMPLETE FOR USE_NEW_STATE

			#if USE_NEW_STATE // propagate jacobian dependencies
			// true idea, but not efficient:
			//subrange(_XNEW_x, iaAbsVel, 0, 16+g_size) += prod(t.VNEW_q, subrange(_XNEW_x, iaOriQuat, 0, 16+g_size));
			//subrange(_XNEW_x, iaPos   , 0, 16+g_size) +=        _dt/2 * subrange(_XNEW_x, iaAbsVel , 0, 16+g_size);

			// efficient actual implementation:
			project(_XNEW_x, iaAbsVel, iaOriQuat) = prod(t.VNEW_q, project(_XNEW_x, iaOriQuat, iaOriQuat));
			project(_XNEW_x, iaAbsVel, iaBiasW  ) = prod(t.VNEW_q, project(_XNEW_x, iaOriQuat, iaBiasW  ));

			project(_XNEW_x, iaPos , iaOriQuat) = _dt/2 * project(_XNEW_x, iaAbsVel, iaOriQuat);
			//project(_XNEW_x,iaPos, iaAbsVel ) = _dt/2 * I // already included above
			project(_XNEW_x, iaPos , iaBiasA  ) = _dt/2 * project(_XNEW_x, iaAbsVel, iaBiasA  );
			project(_XNEW_x, iaPos , iaBiasW  ) = _dt/2 * project(_XNEW_x, iaAbsVel, iaBiasW  );
			project(_XNEW_x, iaPos , iaG      ) = _dt/2 * project(_XNEW_x, iaAbsVel, iaG      );
			#endif



		/* Now on to the perturbation Jacobian XNEW_pert
		 * It is like this:
		 *   var    |  vi             ti             abi   wbi
		 *      pos |  0              3              6     9
		 *   -------+-------------------------------------------------------------------
		 * #if USE_NEW_STATE
		 *   p   0  |I.dt/2  VNEW_q*QNEW_ti*dt/2     0     0     (+ dt/2 * VNEW_)
		 *   q   3  |  0            QNEW_ti          0     0     (QNEW_)
		 *   v   7  |  I       VNEW_q*QNEW_ti        0     0     (+ VNEW_q * QNEW_)
		 * #else
		 *   p   0  |  0              0              0     0
		 *   q   3  |  0            QNEW_ti          0     0
		 *   v   7  |  I              0              0     0
		 * #endif
		 *   ab  10 |  0              0              I     0
		 *   wb  13 |  0              0              0     I
		 *   g   16 |  0              0              0     0
		 */

			// Fill in the easy bits first
			_XNEW_pert.clear();
			project(_XNEW_pert, iaAbsVel, range(0, 3)) = I;
			project(_XNEW_pert, iaBiasA , range(6, 9)) = I;
			project(_XNEW_pert, iaBiasW , range(9, 12)) = I;

			// Tricky bit is QNEW_ti = d(qnew)/d(ti)
			// Here, ti is the integral of the perturbation, ti = integral_{tau=0}^dt (wn(t) * dtau),
			// with: wn: the angular rate measurement noise
			//       dt: the integration period
			//       ti: the resulting angular impulse
			// The integral of the dynamic equation is:
			// q+ = q ** v2q((wm - wb)*dt + ti)
			// We have: QNEW_ti = QNEW_w / dt
			//    with: QNEW_w computed before.
			// The time dependence needs to be included in perturbation.P(), proportional to perturbation.dt:
			//   U = perturbation.P() = U_continuous_time * dt
			//	with: U_continuous_time expressed in ( rad / s / sqrt(s) )^2 = rad^2 / s^3 <-- yeah, it is confusing, but true.
			//   (Use perturbation.set_P_from_continuous() helper if necessary.)
			//
			project(_XNEW_pert, iaOriQuat, range(3, 6)) = t.QNEW_w * (1 / _dt); // QNEW_ti

			#if USE_NEW_STATE // propagate jacobian dependencies
			project(_XNEW_pert, iaAbsVel , range(3, 6)) = prod(t.VNEW_q, project(_XNEW_pert, iaOriQuat, range(3, 6)));

			project(_XNEW_pert, iaPos    , range(0, 6)) = _dt/2 * project(_XNEW_pert, iaAbsVel, range(0, 6)); // all 2 in block
			#endif
		}

#if 1
		vec RobotInertial::e_from_g(const vec3 & _g)
		{
			double norm_g = ublas::norm_2(_g);
			vec3 gf; gf.clear(); gf(2) = -norm_g;
			vec3 rotv = ublasExtra::crossProd(_g,gf);
			double norm_rotv = ublas::norm_2(rotv);
			double sin_a = norm_rotv / (norm_g*norm_g);
			double cos_a = ublas::inner_prod(gf,_g) / (norm_g*norm_g);
			double a = atan2(sin_a, cos_a);
			rotv = a/norm_rotv * rotv;
			return quaternion::q2e(quaternion::v2q(rotv));
		}
#else
		vec RobotInertial::e_from_g(const vec3 & _g)
		{
			vec3 xr, yr, zr, xw, yw, zw; // robot and world frame axis
			xw.clear(); xw(0)=1.; yw.clear(); yw(1)=1.; zw.clear(); zw(2)=1.;
			zr = -_g/ublas::norm_2(_g);
			yr = ublasExtra::crossProd(zr,xw); if (yr(0) < 0.0) yr = -yr;
			xr = -ublasExtra::crossProd(yw,zr); if (xr(0) < 0.0) xr = -xr;
			if (ublas::norm_2(xr) > ublas::norm_2(yr)) // just in case one of them is too close to 0
			{
				xr = xr / ublas::norm_2(xr);
				yr = ublasExtra::crossProd(zr,xr);
			} else
			{
				yr = yr / ublas::norm_2(yr);
				xr = ublasExtra::crossProd(yr,zr);
			}
			mat33 rot;
			rot(0,0)=xr(0); rot(1,0)=xr(1), rot(2,0)=xr(2);
			rot(0,1)=yr(0); rot(1,1)=yr(1), rot(2,1)=yr(2);
			rot(0,2)=zr(0); rot(1,2)=zr(1), rot(2,2)=zr(2);
			vec4 q = q2qc(R2q(rot));
			return q2e(q);
		}
#endif

		void RobotInertial::init_func(const vec & _x, const vec & _u, const vec & _U, vec & _xnew) {
			
			// Separate things out to make it clearer
			vec3 p, v, ab, wb, gv;
			vec g;
			vec4 q;
			splitState(_x, p, q, v, ab, wb, g); // split state vector

			// Split control vector into
			// sensed acceleration and sensed angular rate
			vec3 am, wm;
			splitControl(_u, am, wm);
			
			// init direction and uncertainty of g from acceleration
			const double th_g = 9.81;
			gv = th_g*(-am+ab)/ublas::norm_2(am-ab);
//			double G = (ublas::norm_2(am) - th_g); G = 2*G*G;
//			for (size_t i = pose.size() + 9; i < pose.size() + 9 + g_size; i++){
//				state.P(i,i) = std::max(state.P(i,i), G);
//			}
			if (g_size == 1) { g(0) = -th_g; } else { g = gv; }

			// init orientation from g
			#if INIT_Q_FROM_G
			vec3 e = e_from_g(gv);
			e(2) = q2e(q)(2); // restore initial yaw
			q = e2q(e);
			
			if (g_size == 1) { g(0) = -th_g; } else { g = - z_axis * th_g; } // update g

			vec3 estd; estd.clear();
			if (g_size == 1) // compute attitude initial uncertainty
			{
				vec3 av, wv;
				splitControl(_U, av, wv);

#if 1
				// roughly approximate the euler uncertainty because the exact jacobian is too complicated
				double uncert = 0.;
				int iBiasA = *(iaBiasA.begin());
				for(int i = 0; i < 3; ++i) uncert += av(i)+state.P(iBiasA+i,iBiasA+i);
				uncert = sqrt(uncert/3);
				estd(0) = estd(1) = asin(uncert / 9.81);
#else
				// not working correctly, messes up with yaw and dometimes results in uncertainties too large
				// numerically approximate the euler uncertainty because the exact jacobian is too complicated
				std::cout << "gv " << gv << std::endl;
				vec3 gext;
				for(int i = 0; i < 3; ++i)
				{
					gext = gv;
					std::cout << "uncert " << i << " noise " << sqrt(av(i)) << " bias " << sqrt(state.P(10+i,10+i)) << std::endl;
					gext(i) += sqrt(av(i) + state.P(10+i,10+i)); // uncertainty of measure + bias
					std::cout << "gext " << gext << std::endl;
					vec3 eext = e_from_g(gext);
					std::cout << "eext " << eext << std::endl;
					for(int j = 0; j < 2; ++j) estd(j) = std::max(estd(j), jmath::abs(eext(j)-e(j)));
				}
#endif
				// now convert to quaternion uncertainty
				mat33 E; E.clear(); for(int j = 0; j < 2; ++j) E(j,j) = estd(j)*estd(j);
				mat Q_e(4,3);
				quaternion::e2q_by_de(e, Q_e);
				project(state.P(), iaOriQuat, iaOriQuat) = jmath::ublasExtra::prod_JPJt(E, Q_e);
			}
			#endif
			
			std::cout << "IMU initializes robot state with g = " << g << " (std " << sqrt(state.P(pose.size()+9,pose.size()+9)) <<
				") and orientation q = " << q << " e = " << quaternion::q2e(q) << std::endl;
#if INIT_Q_FROM_G
			if (g_size == 1) std::cout << "Euler attitude is " << e << " (std " << estd << ")" << std::endl;
#endif
			unsplitState(p, q, v, ab, wb, g, _xnew);
		}
		
		
		void RobotInertial::writeLogHeader(kernel::DataLogger& log) const
		{
			std::ostringstream oss; oss << "Robot " << id();
			log.writeComment(oss.str());

			log.writeLegendTokens("time");
			log.writeLegendTokens("sensor");

			log.writeLegendTokens("absx absy absz");
			log.writeLegendTokens("absyaw abspitch absroll");
			log.writeLegendTokens("x y z");
			log.writeLegendTokens("qw qx qy qz");
			log.writeLegendTokens("vx vy vz");
			log.writeLegendTokens("axb ayb azb");
			log.writeLegendTokens("vyawb vpitchb vrollb");
			log.writeLegendTokens("gx gy gz");
			
			log.writeLegendTokens("sig_absx sig_absy sig_absz");
			log.writeLegendTokens("sig_absyaw sig_abspitch sig_absroll");
			log.writeLegendTokens("sig_x sig_y sig_z");
			log.writeLegendTokens("sig_qw sig_qx sig_qy sig_qz");
			log.writeLegendTokens("sig_vx sig_vy sig_vz");
			log.writeLegendTokens("sig_axb sig_ayb sig_azb");
			log.writeLegendTokens("sig_vyawb sig_vpitchb sig_vrollb");
			log.writeLegendTokens("sig_gx sig_gy sig_gz");
		}
		
		void RobotInertial::writeLogData(kernel::DataLogger& log) const
		{
			jblas::vec state_x(6), state_P(6);
			slamPoseToRobotPose(ublas::subrange(state.x(),0,7), ublas::subrange(state.P(),0,7,0,7), state_x, state_P);

			log.writeData(self_time);
			log.writeData(last_updated->name());
			for(int i = 0 ; i < 3 ; ++i) log.writeData(state_x(i));
			for(int i = 0 ; i < 3 ; ++i) log.writeData(state_x(3+2-i));
			for(int i = 0 ; i < 7 ; ++i) log.writeData(state.x()(i));
			for(int i = 7 ; i < 10; ++i) log.writeData(state.x()(i));
			for(int i = 10; i < 13; ++i) log.writeData(state.x()(i));
			for(int i = 13; i < 16; ++i) log.writeData(state.x()(2-(i-13)+13));
			if (g_size == 1) { for(int i = 16; i < 16+2; ++i) log.writeData(0.); log.writeData(state.x()(16)); }
									else { for(int i = 16; i < 16+g_size; ++i) log.writeData(state.x()(i)); }

			
			for(int i = 0 ; i < 3 ; ++i) log.writeData(state_P(i));
			for(int i = 0 ; i < 3 ; ++i) log.writeData(state_P(3+2-i));
			for(int i = 0 ; i < 7 ; ++i) log.writeData(sqrt(state.P()(i,i)));
			for(int i = 7 ; i < 10; ++i) log.writeData(sqrt(state.P()(i,i)));
			for(int i = 10; i < 13; ++i) log.writeData(sqrt(state.P()(i,i)));
			for(int i = 13; i < 16; ++i) log.writeData(sqrt(state.P()(2-(i-13)+13,2-(i-13)+13)));
			if (g_size == 1) { for(int i = 16; i < 16+2; ++i) log.writeData(0.); log.writeData(sqrt(state.P()(16,16))); }
									else { for(int i = 16; i < 16+g_size; ++i) log.writeData(sqrt(state.P()(i,i))); }
		}

	}
}
